import { BINARY_BODY, CORSHeaders, GetContentType, GetRouteResponseContentType, IsValidURL, MimeTypesWithTemplating, ServerErrorCodes, TestHeaderValidity } from '@mockoon/commons';
import cookieParser from 'cookie-parser';
import { EventEmitter } from 'events';
import express from 'express';
import { readFile } from 'fs';
import { createServer as httpCreateServer } from 'http';
import { createProxyMiddleware } from 'http-proxy-middleware';
import { createServer as httpsCreateServer } from 'https';
import killable from 'killable';
import { lookup as mimeTypeLookup } from 'mime-types';
import { basename } from 'path';
import { parse as qsParse } from 'qs';
import { DefaultSSLConfig } from '../constants/ssl.constants';
import { Texts } from '../i18n/en';
import { ResponseRulesInterpreter } from './response-rules-interpreter';
import { TemplateParser } from './template-parser';
import { CreateTransaction } from './utils';
/**
 * Create a server instance from an Environment object.
 *
 * Extends EventEmitter.
 */
export class MockoonServer extends EventEmitter {
    constructor(environment, options = {}) {
        super();
        this.environment = environment;
        this.options = options;
        /**
         * ### Middleware ###
         * Emit the SERVER_ENTERING_REQUEST event
         *
         * @param request
         * @param response
         * @param next
         */
        this.emitEvent = (request, response, next) => {
            this.emit('entering-request');
            next();
        };
        /**
         * ### Middleware ###
         * Add global latency to the mock server
         *
         * @param request
         * @param response
         * @param next
         */
        this.delayResponse = (request, response, next) => {
            this.refreshEnvironment();
            setTimeout(next, this.environment.latency);
        };
        /**
         * ### Middleware ###
         * Parse entering request body
         *
         * @param request
         * @param response
         * @param next
         */
        this.parseBody = (request, response, next) => {
            // Parse body as a raw string and JSON/form if applicable
            const requestContentType = request.header('Content-Type');
            const rawBody = [];
            request.on('data', (chunk) => {
                rawBody.push(Buffer.from(chunk, 'binary'));
            });
            request.on('end', () => {
                request.rawBody = Buffer.concat(rawBody);
                request.body = request.rawBody.toString('utf8');
                try {
                    if (requestContentType) {
                        if (requestContentType.includes('application/json')) {
                            request.bodyJSON = JSON.parse(request.body);
                        }
                        else if (requestContentType.includes('application/x-www-form-urlencoded')) {
                            request.bodyForm = qsParse(request.body, {
                                depth: 10
                            });
                        }
                    }
                }
                catch (error) {
                    this.emit('error', ServerErrorCodes.REQUEST_BODY_PARSE, error);
                }
                next();
            });
        };
        /**
         * ### Middleware ###
         * Emit an event when response emit the 'close' event
         *
         * @param request
         * @param response
         * @param next
         */
        this.logRequest = (request, response, next) => {
            response.on('close', () => {
                this.emit('transaction-complete', CreateTransaction(request, response));
            });
            next();
        };
        /**
         * ### Middleware ###
         * Add environment headers & proxy headers to the response
         *
         * @param request
         * @param response
         * @param next
         */
        this.setResponseHeaders = (request, response, next) => {
            this.setHeaders(this.environment.headers, response, request);
            next();
        };
        /**
         * ### Middleware ###
         * Catch all error handler
         * http://expressjs.com/en/guide/error-handling.html#catching-errors
         *
         * @param server - server on which to log the response
         */
        this.errorHandler = (error, request, response, next) => {
            this.sendError(response, error, 500);
        };
    }
    /**
     * Start a server
     */
    start() {
        const server = express();
        server.disable('x-powered-by');
        server.disable('etag');
        // create https or http server instance
        if (this.environment.https) {
            this.serverInstance = httpsCreateServer(DefaultSSLConfig, server);
        }
        else {
            this.serverInstance = httpCreateServer(server);
        }
        // make serverInstance killable
        this.serverInstance = killable(this.serverInstance);
        // set timeout long enough to allow long latencies
        this.serverInstance.setTimeout(3600000);
        // handle server errors
        this.serverInstance.on('error', (error) => {
            let errorCode = ServerErrorCodes.UNKNOWN_SERVER_ERROR;
            if (error.code === 'EADDRINUSE') {
                errorCode = ServerErrorCodes.PORT_ALREADY_USED;
            }
            else if (error.code === 'EACCES') {
                errorCode = ServerErrorCodes.PORT_INVALID;
            }
            this.emit('error', errorCode, error);
        });
        this.serverInstance.listen(this.environment.port, () => {
            this.emit('started');
        });
        server.use(this.emitEvent);
        server.use(this.delayResponse);
        server.use(this.deduplicateSlashes);
        server.use(cookieParser());
        server.use(this.parseBody);
        server.use(this.logRequest);
        server.use(this.setResponseHeaders);
        this.setRoutes(server);
        this.setCors(server);
        this.enableProxy(server);
        server.use(this.errorHandler);
    }
    /**
     * Kill the server
     */
    stop() {
        if (this.serverInstance) {
            this.serverInstance.kill(() => {
                this.emit('stopped');
            });
        }
    }
    /**
     * ### Middleware ###
     * Remove duplicate slashes in entering call paths
     *
     * @param request
     * @param response
     * @param next
     */
    deduplicateSlashes(request, response, next) {
        request.url = request.url.replace(/\/{2,}/g, '/');
        next();
    }
    /**
     * Generate an environment routes and attach to running server
     *
     * @param server - server on which attach routes
     */
    setRoutes(server) {
        this.environment.routes.forEach((declaredRoute) => {
            // only launch non duplicated routes, or ignore if none.
            if (declaredRoute.enabled) {
                try {
                    const routePath = '/' +
                        (this.environment.endpointPrefix
                            ? this.environment.endpointPrefix + '/'
                            : '') +
                        declaredRoute.endpoint.replace(/ /g, '%20');
                    // create route
                    server[declaredRoute.method](routePath, (request, response) => {
                        // refresh environment data to get route changes that do not require a restart (headers, body, etc)
                        this.refreshEnvironment();
                        const currentRoute = this.getRefreshedRoute(declaredRoute);
                        if (!currentRoute) {
                            this.sendError(response, Texts.EN.MESSAGES.ROUTE_NO_LONGER_EXISTS, 404);
                            return;
                        }
                        const enabledRouteResponse = new ResponseRulesInterpreter(currentRoute.responses, request, currentRoute.randomResponse).chooseResponse();
                        // save route and response UUIDs for logs (only in desktop app)
                        if (declaredRoute.uuid && enabledRouteResponse.uuid) {
                            response.routeUUID = declaredRoute.uuid;
                            response.routeResponseUUID = enabledRouteResponse.uuid;
                        }
                        // add route latency if any
                        setTimeout(() => {
                            const contentType = GetRouteResponseContentType(this.environment, enabledRouteResponse);
                            const routeContentType = GetContentType(enabledRouteResponse.headers);
                            // set http code
                            response.status(enabledRouteResponse.statusCode);
                            this.setHeaders(enabledRouteResponse.headers, response, request);
                            try {
                                // send the file
                                if (enabledRouteResponse.filePath) {
                                    const filePath = TemplateParser(enabledRouteResponse.filePath.replace(/\\/g, '/'), request);
                                    const fileMimeType = mimeTypeLookup(filePath) || '';
                                    // set content-type to route response's one or the detected mime type if none
                                    if (!routeContentType) {
                                        response.set('Content-Type', fileMimeType);
                                    }
                                    if (!enabledRouteResponse.sendFileAsBody) {
                                        response.set('Content-Disposition', `attachment; filename="${basename(filePath)}"`);
                                    }
                                    readFile(filePath, (readError, data) => {
                                        try {
                                            if (readError) {
                                                throw readError;
                                            }
                                            // parse templating for a limited list of mime types
                                            if (MimeTypesWithTemplating.indexOf(fileMimeType) > -1 &&
                                                !enabledRouteResponse.disableTemplating) {
                                                const fileContent = TemplateParser(data.toString(), request);
                                                response.body = fileContent;
                                                response.send(fileContent);
                                            }
                                            else {
                                                response.body = BINARY_BODY;
                                                response.send(data);
                                            }
                                        }
                                        catch (error) {
                                            this.emit('error', ServerErrorCodes.ROUTE_FILE_SERVING_ERROR, error);
                                            this.sendError(response, `${Texts.EN.MESSAGES.ROUTE_FILE_SERVING_ERROR}${error.message}`);
                                        }
                                    });
                                }
                                else {
                                    if (contentType.includes('application/json')) {
                                        response.set('Content-Type', 'application/json');
                                    }
                                    let body = enabledRouteResponse.body;
                                    if (!enabledRouteResponse.disableTemplating) {
                                        body = TemplateParser(body || '', request);
                                    }
                                    response.body = body;
                                    response.send(body);
                                }
                            }
                            catch (error) {
                                this.emit('error', ServerErrorCodes.ROUTE_SERVING_ERROR, error);
                                this.sendError(response, `${Texts.EN.MESSAGES.ROUTE_SERVING_ERROR}${error.message}`);
                            }
                        }, enabledRouteResponse.latency);
                    });
                }
                catch (error) {
                    let errorCode = ServerErrorCodes.ROUTE_CREATION_ERROR;
                    // if invalid regex defined
                    if (error.message.indexOf('Invalid regular expression') > -1) {
                        errorCode = ServerErrorCodes.ROUTE_CREATION_ERROR_REGEX;
                    }
                    this.emit('error', errorCode, error);
                }
            }
        });
    }
    /**
     * Always answer with status 200 to CORS pre flight OPTIONS requests if option activated.
     * /!\ Must be called after the routes creation otherwise it will intercept all user defined OPTIONS routes.
     *
     * @param server - express instance
     */
    setCors(server) {
        if (this.environment.cors) {
            server.options('/*', (req, res) => {
                this.refreshEnvironment();
                // override default CORS headers with environment's headers
                this.setHeaders([...CORSHeaders, ...this.environment.headers], res, req);
                res.status(200).end();
            });
        }
    }
    /**
     * Add catch-all proxy if enabled.
     * Restream the body to the proxied API because it already has been
     * intercepted by the body parser.
     *
     * @param server - server on which to launch the proxy
     */
    enableProxy(server) {
        if (this.environment.proxyMode &&
            this.environment.proxyHost &&
            IsValidURL(this.environment.proxyHost)) {
            this.emit('creating-proxy');
            server.use('*', createProxyMiddleware({
                target: this.environment.proxyHost,
                secure: false,
                changeOrigin: true,
                logProvider: this.options.logProvider,
                ssl: Object.assign(Object.assign({}, DefaultSSLConfig), { agent: false }),
                onProxyReq: (proxyReq, request, response) => {
                    this.refreshEnvironment();
                    request.proxied = true;
                    this.setHeaders(this.environment.proxyReqHeaders, proxyReq, request);
                    // re-stream the body (intercepted by body parser method)
                    if (request.rawBody) {
                        proxyReq.write(request.rawBody);
                    }
                },
                onProxyRes: (proxyRes, request, response) => {
                    this.refreshEnvironment();
                    let body = '';
                    proxyRes.on('data', (chunk) => {
                        body += chunk;
                    });
                    proxyRes.on('end', () => {
                        response.body = body;
                    });
                    this.setHeaders(this.environment.proxyResHeaders, proxyRes, request);
                },
                onError: (error, request, response) => {
                    this.emit('error', ServerErrorCodes.PROXY_ERROR, error);
                    this.sendError(response, `${Texts.EN.MESSAGES.PROXY_ERROR}${this.environment.proxyHost}${request.url}: ${error}`, 504);
                }
            }));
        }
    }
    /**
     * Set the provided headers on the target. Use different headers accessors
     * depending on the type of target:
     * express.Response/http.OutgoingMessage/http.IncomingMessage
     * Use the source in the template parsing of each header value.
     *
     * @param headers
     * @param target
     * @param source
     */
    setHeaders(headers, target, source) {
        headers.forEach((header) => {
            if (header.key && header.value && !TestHeaderValidity(header.key)) {
                let parsedHeaderValue;
                try {
                    parsedHeaderValue = TemplateParser(header.value, source);
                }
                catch (error) {
                    const errorMessage = Texts.EN.MESSAGES.HEADER_PARSING_ERROR;
                    parsedHeaderValue = errorMessage;
                }
                if (target.set) {
                    // for express.Response
                    target.set(header.key, parsedHeaderValue);
                }
                else if (target.setHeader) {
                    // for proxy http.OutgoingMessage
                    target.setHeader(header.key, parsedHeaderValue);
                }
                else {
                    // for http.IncomingMessage
                    target.headers[header.key] = parsedHeaderValue;
                }
            }
        });
    }
    /**
     * Send an error with text/plain content type, the provided message and status code.
     * Status is optional. No status will default to the one defined by the user, allowing for maximum customization.
     *
     * @param response
     * @param errorMessage
     * @param status
     */
    sendError(response, errorMessage, status) {
        response.set('Content-Type', 'text/plain');
        response.body = errorMessage;
        if (status !== undefined) {
            response.status(status);
        }
        response.send(errorMessage);
    }
    /**
     * Request an updated environment to allow
     * modification of some parameters without a restart (latency, headers, etc)
     */
    refreshEnvironment() {
        if (this.options.refreshEnvironmentFunction && this.environment.uuid) {
            const updatedEnvironment = this.options.refreshEnvironmentFunction(this.environment.uuid);
            if (updatedEnvironment) {
                this.environment = updatedEnvironment;
            }
        }
    }
    /**
     * Request an updated route to allow
     * modification of some parameters without a restart (latency, headers, etc)
     * This only makes sense if the refreshEnvironmentFunction has been provided.
     *
     * @param routeUUID
     */
    getRefreshedRoute(currentRoute) {
        if (this.options.refreshEnvironmentFunction && this.environment.uuid) {
            return this.environment.routes.find((route) => route.uuid === currentRoute.uuid);
        }
        return currentRoute;
    }
}
//# sourceMappingURL=server.js.map